"""Multi-agents."""

from neuroagent.multi_agents.base_multi_agent import BaseMultiAgent
from neuroagent.multi_agents.supervisor_multi_agent import SupervisorMultiAgent

__all__ = [
    "BaseMultiAgent",
    "SupervisorMultiAgent",
]
